/*
 * tmos - i51.c
 *
 * SSP for i51 serial
 * 
 * Copyright (C) 2010, 2011  Krzysztof Mazur <krzysiek@podlesie.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <tmos/kernel.h>
#include <tmos/serial.h>
#include <tmos/port.h>
#include <tmos/crc16.h>
#include <asm/8052.h>
#include <asm/system.h>
#include <asm/task.h>
#include <tmos/errno.h>
#include <tmos/net.h>
#include <asm/irqs.h>
#include <limits.h>

#ifndef dir_tx
#define dir_tx()	do { } while (0)
#endif
#ifndef dir_rx
#define dir_rx()	do { } while (0)
#endif

#define serial_interrupt_disable() do { ES = 0; } while (0)
#define serial_interrupt_enable() do { ES = 1; } while (0)

#define SSP_MTU		((CONFIG_SSP_MTU + 6) / 7)

#define TXP_END 0xff

#ifdef CONFIG_KERNEL_DATA
static uint8_t __idata msg_buffer[8 * SSP_MTU];
#else
static uint8_t __kernel msg_buffer[8 * SSP_MTU];
#endif
static uint8_t __kernel msg_size = 0;
static uint8_t __kernel msg_txp = TXP_END;	/* transmit pointer */
static __bit rx_lock = 0;
static __bit esc = 0;
uint8_t __kernel ssp_errors = 0;

#ifdef CONFIG_SSP_ADDR_FILTER

#ifdef CONFIG_SSP_ADDR_FILTER_RUNTIME
static uint8_t ssp_addr = CONFIG_SSP_ADDR_FILTER_ADDR;
#define ssp_address()	ssp_addr
#else
#define ssp_address()	CONFIG_SSP_ADDR_FILTER_ADDR
#endif

#ifdef CONFIG_SSP_ADDR_FILTER_BROADCAST
#define msg_filter() \
	((msg_buffer[0] == ssp_address()) || !msg_buffer[0])
#else
#define msg_filter() (msg_buffer[0] == ssp_address())
#endif
#else
#define msg_filter() (1)
#endif

#if CONFIG_SSP_1A9
#define tx_send_ssp_data(x) \
	do { \
		TB8 = 0; \
		SBUF = x; \
	} while (0)

#define tx_send_ssp_eof(x) \
	do { \
		TB8 = 1; \
		SBUF = x; \
	} while (0)

#elif CONFIG_SSP_1A8

#define __tx_send_ssp_data(x) do { SBUF = (x); } while (0)
#define __tx_send_ssp_eof(x) do { SBUF = (x) | 0x80; } while (0)
#define tx_send_ssp_data(x) __tx_send_ssp_data(x)
#define tx_send_ssp_eof(x) __tx_send_ssp_eof(x)
#else
#error "unsupported version"
#endif

void ser_irq(void) __interrupt SERIAL_INTERRUPT __using I51_SERIAL_BANK
{
	unsigned char tmp;

	if (RI) {
		RI = 0;
		tmp = SBUF;

		if (!rx_lock) {
			if (msg_size < 8 * SSP_MTU) {
				msg_buffer[msg_size] = tmp;
				msg_size++;
			} else {
				msg_size = 8 * SSP_MTU + 1;
			}

			if (tmp & 0x80) {
				if (((uint8_t) (msg_size - 2)
							<= 8 * SSP_MTU - 2)
						&& msg_filter()) {
					msg_txp = TXP_END;
					rx_lock = 1;
				} else {
					msg_size = 0;
				}
			}
		}
	}

	if (TI) {
		TI = 0;
		if ((uint8_t)(msg_txp + 1) < msg_size) {
			__tx_send_ssp_data(msg_buffer[msg_txp]);
			msg_txp++;
		} else if ((uint8_t) (msg_txp + 1) == msg_size) {
			__tx_send_ssp_eof(msg_buffer[msg_txp]);
			msg_txp++;
		} else {
			dir_rx();
			msg_size = 0;
			rx_lock = 0;
		}
	}
}

char ssp_i51_ioctl(unsigned char cmd, void __kernel *arg) __reentrant
{
	char ret = -EINVAL;

	switch (cmd) {
#ifdef CONFIG_SSP_ADDR_FILTER_RUNTIME
	case NET_IOCTL_SET_ADDR_BYTE:
		ssp_addr = *((uint8_t __kernel *) arg);
		ret = 0;
		break;
#endif
	case NET_IOCTL_SET_SPEED:
		ret = 0;
		switch (*((uint8_t __kernel *) arg)) {
		case NET_SPEED_9600:
			ser_set_spd(B9600);
		default:
			ret = -EINVAL;
		}
		break;
	}
	return ret;
}

#if CONFIG_SSP_1A8
/**
 * ssp_decode - convert SSP-1A 8-bit block to 8-bit data
 * @src: source block
 * @dst: 
 *
 * The src and dst may alias.
 */
static void ssp_decode(unsigned char __kernel *src, unsigned char __kernel *dst)
{
	unsigned char __kernel *s = src;
	unsigned char __kernel *d = dst;

	d[0] = (s[1] << 7) | (s[0] >> 0);
	d[1] = (s[2] << 6) | (s[1] >> 1);
	d[2] = (s[3] << 5) | (s[2] >> 2);
	d[3] = (s[4] << 4) | (s[3] >> 3);
	d[4] = (s[5] << 3) | (s[4] >> 4);
	d[5] = (s[6] << 2) | (s[5] >> 5);
	d[6] = (s[7] << 1) | (s[6] >> 6);
}

/**
 * ssp_encode - convert 8-bit data to SSP-1A 8-bit block
 * @src: source block
 * @dst: 
 *
 * The src and dst may alias.
 */
static void ssp_encode(unsigned char __kernel *src, unsigned char __kernel *dst)
{
	unsigned char __kernel *s = src;
	unsigned char __kernel *d = dst;

	d[7] = (s[6] >> 1);
	d[6] = ((s[6] << 6) | (s[5] >> 2)) & 0x7f;
	d[5] = ((s[5] << 5) | (s[4] >> 3)) & 0x7f;
	d[4] = ((s[4] << 4) | (s[3] >> 4)) & 0x7f;
	d[3] = ((s[3] << 3) | (s[2] >> 5)) & 0x7f;
	d[2] = ((s[2] << 2) | (s[1] >> 6)) & 0x7f;
	d[1] = ((s[1] << 1) | (s[0] >> 7)) & 0x7f;
	d[0] = s[0] & 0x7f;
}

/*
 * @len: frame length, must be > 0
 */
static unsigned char ssp_1a_encode(unsigned char __kernel *buf,
		unsigned char len) __reentrant
{
	unsigned char out = (len / 7) * 8;
	unsigned char __kernel *d = buf + (len / 7) * 8;
	unsigned char __kernel *end = buf;
	unsigned char __kernel *s = buf + (len / 7) * 7;

	if (len)
		out += (len % 7) + 1;

	for (;;) {
		ssp_encode(s, d);
		d -= 8;
		s -= 7;
		if (s < end)
			break;
	}
	return out;
}

static int ssp_1a_decode(unsigned char __kernel *buf, unsigned char len)
	__reentrant
{
	unsigned char out = 0;
	unsigned char __kernel *d = buf;
	unsigned char __kernel *end = buf + len;
	unsigned char __kernel *s = buf;

	out = 7 * (len / 8) + (len & 7) - 1;
	for (;;) {
		ssp_decode(s, d);
		s += 8;
		d += 7;
		if (s >= end)
			break;
	}
	return out;
}
#endif

void ssp_i51_poll(void) __reentrant
{
	if (rx_lock && (msg_txp == TXP_END)) {
		net_size_t ret = 0;
		uint16_t __kernel *crc;
		struct net_msg msg;

		/* setup message */
		msg_init(&msg);
		msg_set_interface(&msg, NET_IFACE_SERIAL0);
		msg_set_buffer(&msg, (void __kernel *)msg_buffer);
		msg_set_head(&msg, 0);

#if CONFIG_SSP_1A8
		msg_size = ssp_1a_decode((void __kernel *) msg_buffer,
				msg_size);
#endif

#if CONFIG_SSP_CRC16
		msg_set_tail(&msg, msg_size - 2);
		msg_set_size(&msg, 7 * SSP_MTU - 2);
		crc = (uint16_t __kernel *)(msg_buffer + msg_size - 2);
		if (le16_to_cpu(*crc) == crc16_kernel(0xffff,
					(void __kernel *)msg_buffer,
					msg_size - 2)) {
			ret = net_xmit(NET_PROTO_MODBUS, &msg);
			if (ret) {
				msg_size = ret + 2;
				crc = (uint16_t __kernel *)(msg_buffer + ret);
				*crc = cpu_to_le16(crc16_kernel(0xffff,
							(void __kernel *)
							msg_buffer, ret));
			}
		} else {
			ssp_errors++; 
		}
#else
		msg_set_tail(&msg, msg_size);
		msg_set_size(&msg, 7 * SSP_MTU);
		ret = net_xmit(NET_PROTO_MODBUS, &msg);
		if (ret)
			msg_size = ret;
#endif
#if CONFIG_SSP_1A8
		msg_size = ssp_1a_encode((void __kernel *) msg_buffer,
				msg_size);
#endif

		serial_interrupt_disable();
		if (ret) {
			dir_tx();
			msg_txp = 0;

			/*
			 * It's a good idea to mark also beginning of frame,
			 * especially for some broken RS-485 links without
			 * bias that can generate some start bits.
			 *
			 * This causes generation of frame with start
			 * bit and maximal number of ones in case of
			 * no parity, mark parity or odd parity.
			 * The receiver will be properly synchronized
			 * after this number of ones.
			 */
			tx_send_ssp_eof(0x7f);
		} else {
			msg_size = 0;
			rx_lock = 0;
		}
		serial_interrupt_enable();
	}
}

static void ssp_i51_init(void) __reentrant
{
	dir_rx();
}

device_initcall(ssp_i51_init)
