/*
 * tmos - system.h
 *
 * Copyright (C) 2009  Krzysztof Mazur <krzysiek@podlesie.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef __ASM_SYSTEM_H
#define __ASM_SYSTEM_H

#include <asm/barrier.h>
#include <asm/processor.h>
#include <asm/8052.h>
#include <tmos/config.h>

/*
 * We use one register bank per context.
 */
#define SOFTIRQ_BANK		1
#define HIGH_PRIORITY_BANK	2
#define LOW_PRIORITY_BANK	3

#define setup_registers(bank) \
	__asm__(ar0 = (bank << 3) + 0x00); \
	__asm__(ar1 = (bank << 3) + 0x01); \
	__asm__(ar2 = (bank << 3) + 0x02); \
	__asm__(ar3 = (bank << 3) + 0x03); \
	__asm__(ar4 = (bank << 3) + 0x04); \
	__asm__(ar5 = (bank << 3) + 0x05); \
	__asm__(ar6 = (bank << 3) + 0x06); \
	__asm__(ar7 = (bank << 3) + 0x07);

typedef unsigned char cpu_flags_t;

#define local_save_flags(x) \
		do { \
			typecheck(cpu_flags_t, x); \
			(x) = EA; __acquire(IRQ); \
		} while (0)

#define local_irq_restore(x) \
	do { \
		typecheck(cpu_flags_t, x); \
		__release(IRQ); \
		EA = (x); \
	} while (0)
#define local_irq_disable()	do { EA = 0; __acquire(IRQ); } while (0)
#define local_irq_enable()	do { EA = 1; __release(IRQ); } while (0)
#define local_irq_save(x)						\
	do {								\
		typecheck(cpu_flags_t, x);				\
		x = EA;							\
		local_irq_disable();					\
	} while (0)


#if CONFIG_PREEMPT
extern char preempt_count;

#define preempt_disable() \
	do { \
		preempt_count++; \
		__acquire(PREEMPT); \
		barrier(); \
	} while (0)

#define preempt_enable() \
	do { \
		barrier(); \
		__release(PREEMPT); \
		preempt_count--; \
	} while (0)

#define preempt_enabled() (!preempt_count)
#define preempt_enabled_flags(flags) (!preempt_count)

#else

#define preempt_disable()		__acquire(PREEMPT)
#define preempt_enable()		__release(PREEMPT)
#define preempt_enabled()		(0)
#define preempt_enabled_flags(flags)	(0)
#endif

#define __kcode __code

#if CONFIG_KERNEL_DATA
#define __kernel __data
typedef uint8_t kernel_size_t;
#define KERNEL_TASK_SIZE	256
#else
#define __kernel
typedef size_t kernel_size_t;
#define KERNEL_TASK_SIZE	(1 << 24UL)
#endif

#if CONFIG_HIGHMEM
#define __highmem __xdata
#else
#define __highmem __data
#endif

#define __kernel __data

#if CONFIG_PREEMPT
#define __preempt __reentrant
#else
#define __preempt
#endif

#define copy_to_user(dst, src, len)	(memcpy(dst, src, len), 0)
#define copy_from_user(dst, src, len)	(memcpy(dst, src, len), 0)

void arch_init(void);

#ifdef BIG_ENDIAN
#define set16(dst, high, low) \
	do { \
		((uint8_t *)(dst))[0] = (high); \
		((uint8_t *)(dst))[1] = (low); \
	} while (0)
#else
#define set16(dst, high, low) \
	do { \
		((uint8_t *)(dst))[1] = (high); \
		((uint8_t *)(dst))[0] = (low); \
	} while (0)
#endif

#if CONFIG_PAGE_ALLOC
#define PAGE_SHIFT	7
#define PAGE_SIZE	128
#define PAGE_NULL	(0xff)

#define NUM_PAGES	(CONFIG_XDATA_SIZE / PAGE_SIZE)

typedef unsigned char page_t;

page_t page_alloc(void) __reentrant;
void page_free(page_t) __reentrant;

#define page_to_virt(p)	((void __xdata *) ((unsigned int)(p) << PAGE_SHIFT))
#define virt_to_page(p) ((page_t) ((unsigned int)(p) >> PAGE_SHIFT))
#endif

#if CONFIG_TDATA
#define TDATA_BASE	0x80
#endif

#define put_user(x, ptr)	do { ((*ptr) = (x)); } while (0)
#define get_user(x, ptr)	do { ((x) = (*ptr)); } while (0)

#ifdef __INLINE_ASM
#define __init_call(stage, x) \
	static void __ ## stage ## init_call(void) __naked \
	{ \
		__asm \
			.area stage ##_initcall_sec(CODE,REL,CON) \
			.db _##x \
			.db _##x >> 8\
		__endasm; \
	}
#else
#define __init_call(stage, x) \
	void __ ## x(void); \
	void __ ## x(void) \
	{ \
		x(); \
	}
#endif

#define arch_initcall(x) __init_call(arch, x)
#define subsys_initcall(x) __init_call(subsys, x)
#define device_initcall(x) __init_call(device, x)
#define late_initcall(x) __init_call(late, x)

enum stage {
	ARCH_INIT,
	SUBSYS_INIT,
	DEVICE_INIT,
	LATE_INIT
};

void call_initcalls(enum stage stage) __reentrant;

#define ptr_is_code(ptr) \
	((((unsigned long) ptr) & 0x800000UL) ? 1 : 0)
#define ptr_is_highmem(ptr) \
	((((unsigned long) ptr) & 0x400000UL) ? 1 : 0)
#define ptr_is_data(ptr) \
	((((unsigned long) ptr) & 0xff0000UL) ? 0 : 1)

#include <tmos/string.h>

#endif
